local lg = love.graphics
local tinsert = table.insert

local leafIcon = lg.newImage("leaf.png")
local fireIcon = lg.newImage("circle2.png")
local magicIcon = lg.newImage("plus2.png")
local floor = math.floor

local particle = {}
particle.__index = particle


function getPS(name, image)
    local ps_data = require(name)
    local particle_settings = {}
    particle_settings["colors"] = {}
    particle_settings["sizes"] = {}
    for k, v in pairs(ps_data) do
        if k == "colors" then
            local j = 1
            for i = 1, #v , 4 do
                local color = {v[i], v[i+1], v[i+2], v[i+3]}
                particle_settings["colors"][j] = color
                j = j + 1
            end
        elseif k == "sizes" then
            for i = 1, #v do particle_settings["sizes"][i] = v[i] end
        else particle_settings[k] = v end
    end
    local ps = love.graphics.newParticleSystem(image, particle_settings["buffer_size"])
    ps:setAreaSpread(string.lower(particle_settings["area_spread_distribution"]), particle_settings["area_spread_dx"] or 0 , particle_settings["area_spread_dy"] or 0)
    ps:setBufferSize(particle_settings["buffer_size"] or 1)
    local colors = {}
    for i = 1, 8 do 
        if particle_settings["colors"][i][1] ~= 0 or particle_settings["colors"][i][2] ~= 0 or particle_settings["colors"][i][3] ~= 0 or particle_settings["colors"][i][4] ~= 0 then
            table.insert(colors, particle_settings["colors"][i][1] or 0)
            table.insert(colors, particle_settings["colors"][i][2] or 0)
            table.insert(colors, particle_settings["colors"][i][3] or 0)
            table.insert(colors, particle_settings["colors"][i][4] or 0)
        end
    end
    ps:setColors(unpack(colors))
    ps:setColors(unpack(colors))
    ps:setDirection(math.rad(particle_settings["direction"] or 0))
    ps:setEmissionRate(particle_settings["emission_rate"] or 0)
    ps:setEmitterLifetime(particle_settings["emitter_lifetime"] or 0)
    ps:setInsertMode(string.lower(particle_settings["insert_mode"]))
    ps:setLinearAcceleration(particle_settings["linear_acceleration_xmin"] or 0, particle_settings["linear_acceleration_ymin"] or 0, 
                             particle_settings["linear_acceleration_xmax"] or 0, particle_settings["linear_acceleration_ymax"] or 0)
    if particle_settings["offsetx"] ~= 0 or particle_settings["offsety"] ~= 0 then
        ps:setOffset(particle_settings["offsetx"], particle_settings["offsety"])
    end
    ps:setParticleLifetime(particle_settings["plifetime_min"] or 0, particle_settings["plifetime_max"] or 0)
    ps:setRadialAcceleration(particle_settings["radialacc_min"] or 0, particle_settings["radialacc_max"] or 0)
    ps:setRotation(math.rad(particle_settings["rotation_min"] or 0), math.rad(particle_settings["rotation_max"] or 0))
    ps:setSizeVariation(particle_settings["size_variation"] or 0)
    local sizes = {}
    local sizes_i = 1 
    for i = 1, 8 do 
        if particle_settings["sizes"][i] == 0 then
            if i < 8 and particle_settings["sizes"][i+1] == 0 then
                sizes_i = i
                break
            end
        end
    end
    if sizes_i > 1 then
        for i = 1, sizes_i do table.insert(sizes, particle_settings["sizes"][i] or 0) end
        ps:setSizes(unpack(sizes))
    end
    ps:setSpeed(particle_settings["speed_min"] or 0, particle_settings["speed_max"] or 0)
    ps:setSpin(math.rad(particle_settings["spin_min"] or 0), math.rad(particle_settings["spin_max"] or 0))
    ps:setSpinVariation(particle_settings["spin_variation"] or 0)
    ps:setSpread(math.rad(particle_settings["spread"] or 0))
    ps:setTangentialAcceleration(particle_settings["tangential_acceleration_min"] or 0, particle_settings["tangential_acceleration_max"] or 0)
    return ps
end

function particle:Die()
	if not self.dying then
		self.dying = true
		self.timetodie = 1
		for k,v in pairs(self.particles) do
			v.ps:stop()
		end
	end
end

function particle:UpdatePositionX(dt)
	self.newx = self.x + self.vx*dt
	self.xCheck = self.vx > 0 and self.newx+self.radius or self.newx-self.radius

	if self.canCollide then
		for j = floor(self.y-self.radius), floor(self.y+self.radius), self.radius do
			if self.world:Collides(self.xCheck, j) then
				self:Die()
				return
			end
		end
	end
		
	self.x = self.newx
end

function particle:UpdatePositionY(dt)
	self.newy = self.y + self.vy*dt
	self.yCheck = self.vy > 0 and self.newy+self.radius or self.newy-self.radius

	if self.canCollide then
		for i = floor(self.x-self.radius), floor(self.x+self.radius), self.radius do
			if self.world:Collides(i, self.yCheck) then
				self:Die()
				return
			end
		end
	end
	self.y = self.newy
end

function particle:UpdatePosition(dt)
	if self:HasGravity() then
		self.vy = self.vy + self.gravity*dt
	end
	if not self.dying then
		self:UpdatePositionX(dt)
		self:UpdatePositionY(dt)
	end
end

function particle:UpdateLifetime(dt)
	if self:Decays() then
		self.lifetime = self.lifetime - dt
		if self.lifetime < 0 and not self.dying then
			self:Die()
		end
	end
end

function particle:Update(dt)
	self.rot = self.rot+self.vrot*dt

	self:UpdateLifetime(dt)
	self:UpdateParticles(dt)
	self:UpdateDeath(dt)
	self:UpdatePosition(dt)

end

function particle:UpdateDeath(dt)
	if self.dying then
		self.timetodie = self.timetodie - dt
		if self.timetodie < 0 then
			self:Delete()
		end
	end
end

function particle:Delete()
	DeleteSpell(self)
end

function particle:Decays()
	return self.decays
end

function particle:HasGravity()
	return self.hasGravity
end

function particle:Draw()
	lg.setColor(unpack(self.color))
	--lg.draw(self.img, self.x,self.y,self.rot,self.scale,self.scale,self.radius/2,self.radius/2)
	self:DrawParticles()

	--self:DebugDraw()
end

function particle:DebugDraw()
	for i = floor(self.x-self.radius), floor(self.x+self.radius), self.radius do
		for j = floor(self.y-self.radius), floor(self.y +self.radius), self.radius do
			lg.rectangle("fill", i, j, 1, 1)
		end
	end
end

function particle:StartParticles()
	for k,p in pairs(self.particles) do
		p.ps:start()
	end
end

function particle:UpdateParticles(dt)
	for k,p in pairs(self.particles) do
		p.ps:update(dt)
		if not p.isCore then
			p.ps:moveTo(self.x, self.y)
		end
	end
end

function particle:DrawParticles()
	for k,p in pairs(self.particles) do
		if p.blendMode then
			lg.setBlendMode(p.blendMode)
		end
		if p.isCore then
			lg.draw(p.ps, self.x,self.y)
		else
			lg.draw(p.ps, 0,0)
		end
		lg.setBlendMode("alpha")
	end
end

function particle:ParseRecipe(recipe)
	if recipe.prop.life > 0 then
		local v = recipe.prop.life
		local core = getPS("leafcore5", leafIcon)
		core:setEmissionRate((200/5)*v)
		core:setTangentialAcceleration(50+v*40, 50+v*40)
		tinsert(self.particles, {isCore = true, ps = core, priority = v})
		local trail = getPS("leaf", leafIcon)
		trail:setEmissionRate(v*5)
		tinsert(self.particles, {isCore = false, ps = trail, priority = v})
	end

	if recipe.prop.fire > 0 then
		local v = recipe.prop.fire
		local core = getPS("FireCore", fireIcon)
		--core:setAreaSpread("none", 0, 0)
		--ecore:setSpeed(v*3)
		local trail = getPS("FireTrail", fireIcon)
		trail:setAreaSpread("normal", 3+v, 3+v)
		trail:setEmissionRate(200+v*100)
		tinsert(self.particles, {isCore = true, ps = core, blendMode = "additive", priority = v})
		tinsert(self.particles, {isCore = false, ps = trail, blendMode = "additive", priority = v})
	end

	if recipe.prop.magic > 0 then
		local v = recipe.prop.magic
		local core = getPS("Magic3", magicIcon)
		tinsert(self.particles, {isCore = true, ps = core, blendMode = "additive", priority = v})
	end

	if recipe.prop.earth > 0 then
		local v = recipe.prop.earth
		self.hasGravity = true
		
		-- more earth = heavier but more magic = further
		self.gravity = math.min(v*200, 1000- (recipe.prop.magic*200))
	end

	if recipe.prop.magic < 5 then
		self.canCollide = true
	end

	table.sort(self.particles, function(a,b) return a.priority < b.priority end)
end

function NewParticle(world, recipe,parent, x,y,angle)
	local s = setmetatable({}, particle)
	s.x = x
	s.y = y
	s.radius = 15
	s.speed = 300
	s.vx = math.sin(angle)*s.speed-- hyp = 10, we want opp  sin(angle) = opp over hyp
	s.vy = math.cos(angle)*s.speed
	s.rot = 0
	s.vrot = 10
	s.img = lg.newImage("circle.jpg")
	s.color = {255,255,255,255}
	s.scale = 0.5
	s.radius = 10
	s.gravity = 200
	s.lifetime = 5
	s.parent = parent
	s.decays = true 
	s.recipe = recipe
	s.world = world
	s.particles = {}
	s:ParseRecipe(recipe)
	parent[s] = s
	s:StartParticles()

	return s
end


